
/*
 * Copyright (C) 2024, Advanced Micro Devices. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS!
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "alcp/alcp.h"
#include <inttypes.h>
#include <malloc.h>
#include <stdio.h>

#define RED   "\033[0;31m"
#define RESET "\033[0m"

static alc_drbg_handle_t handle;

void
printHexString(Uint8* bytes, int length)
{
    size_t i;
    for (i = 0; i < length; i++) {
        printf("%02x", bytes[i]);
    }
}

alc_error_t
test_drbg(alc_drbg_info_t drbg_info,
          Uint8*          personalizationString,
          Uint64          personalizationStringLen,
          Uint8*          AdditionalInput1,
          Uint64          AdditionalInputLength1,
          Uint8*          AdditionalInput2,
          Uint64          AdditionalInputLength2)
{
    alc_error_t err = ALC_ERROR_NONE;
    err             = alcp_drbg_supported(&drbg_info);

    if (!alcp_is_error(err)) {
        handle.ch_context = malloc(alcp_drbg_context_size(&drbg_info));
    } else {
        printf("DRBG Information provided is unsupported\n");
        return err;
    }

    err = alcp_drbg_request(&handle, &drbg_info);
    if (alcp_is_error(err)) {
        printf("Error Occurred on DRBG Request - %10" PRId64 "\n", err);
        return err;
    }
    const int cSecurityStrength = 100;
    err                         = alcp_drbg_initialize(&handle,
                               cSecurityStrength,
                               personalizationString,
                               personalizationStringLen);
    if (alcp_is_error(err)) {
        printf("Error Occurred on DRBG initialize - %10" PRId64 "\n", err);
        return err;
    }
    Uint8 output[64];
    err = alcp_drbg_randomize(&handle,
                              output,
                              sizeof(output),
                              cSecurityStrength,
                              AdditionalInput1,
                              AdditionalInputLength1);
    if (alcp_is_error(err)) {
        printf("Error Occurred on DRBG initialize - %10" PRId64 "\n", err);
        return err;
    }
    printf("First Call: Randomly generated bytes: \n");
    printHexString(output, sizeof(output));
    printf("\n");
    // Generating Again
    err = alcp_drbg_randomize(&handle,
                              output,
                              sizeof(output),
                              cSecurityStrength,
                              AdditionalInput2,
                              AdditionalInputLength2);
    if (alcp_is_error(err)) {
        printf("Error Occurred on DRBG initialize - %10" PRId64 "\n", err);
        return err;
    }
    printf("Second Call: Randomly generated bytes: \n");
    printHexString(output, sizeof(output));
    printf("\n");

    alcp_drbg_finish(&handle);

    if (handle.ch_context) {
        free(handle.ch_context);
        handle.ch_context = NULL;
    }
    return err;
}

alc_error_t
test_ctr_drbg()
{
    Uint8 customEntropy[] = { 0xca, 0xe4, 0x8d, 0xd8, 0x0d, 0x29, 0x81, 0x03,
                              0xef, 0x1e, 0xc0, 0xbf, 0x1b, 0xb9, 0x62, 0x70 };

    Uint8 nonce[] = { 0xd8, 0x27, 0xf9, 0x16, 0x13, 0xe0, 0xb4, 0x7f };
    Uint8 personalization_string[] = { 0xcc, 0x92, 0x8f, 0x3d, 0x2d, 0xf3,
                                       0x1a, 0x29, 0xf4, 0xe4, 0x44, 0xf3,
                                       0xdf, 0x08, 0xbe, 0x21 };
    Uint8 AdditionalInput1[]       = {
        0x7e, 0xaa, 0x1b, 0xbe, 0xc7, 0x93, 0x93, 0xa7,
        0xf4, 0xa8, 0x22, 0x7b, 0x69, 0x1e, 0xcb, 0x68
    };
    Uint8 AdditionalInput2[] = {
        0x68, 0x69, 0xc6, 0xc7, 0xb9, 0xe6, 0x65, 0x3b,
        0x39, 0x77, 0xf0, 0x78, 0x9e, 0x94, 0x47, 0x8a
    };

    alc_drbg_info_t
        drbg_info = { .di_type           = ALC_DRBG_CTR,
                      .max_entropy_len   = sizeof(customEntropy),
                      .max_nonce_len     = sizeof(nonce),
                      .di_algoinfo       = { .ctr_drbg = { .di_keysize = 128,
                                                     .use_derivation_function =
                                                         true } },
                      .di_rng_sourceinfo = {
                          .custom_rng    = true,
                          .di_sourceinfo = {
                              .custom_rng_info = {
                                  .entropy    = customEntropy,
                                  .entropylen = sizeof(customEntropy),
                                  .nonce      = nonce,
                                  .noncelen   = sizeof(nonce) } } } };

    alc_error_t ret = test_drbg(drbg_info,
                                personalization_string,
                                sizeof(personalization_string),
                                AdditionalInput1,
                                sizeof(AdditionalInput1),
                                AdditionalInput2,
                                sizeof(AdditionalInput2));
    return ret;
}

alc_error_t
test_hmac_drbg()
{
    Uint8 customEntropy[] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
                              0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
                              0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
                              0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
                              0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
                              0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
                              0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };

    Uint8 nonce[] = { 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27 };

    alc_drbg_info_t
        drbg_info = { .di_type           = ALC_DRBG_HMAC,
                      .max_entropy_len   = sizeof(customEntropy),
                      .max_nonce_len     = sizeof(nonce),
                      .di_algoinfo       = { .hmac_drbg = { .digest_mode =
                                                          ALC_SHA2_256 } },
                      .di_rng_sourceinfo = {
                          .custom_rng    = true,
                          .di_sourceinfo = {
                              .custom_rng_info = {
                                  .entropy    = customEntropy,
                                  .entropylen = sizeof(customEntropy),
                                  .nonce      = nonce,
                                  .noncelen   = sizeof(nonce) } } } };

    alc_error_t ret = test_drbg(drbg_info, NULL, 0, NULL, 0, NULL, 0);
    return ret;
}
int
main(int argc, char const* argv[])
{
    alc_error_t err = ALC_ERROR_NONE;
    printf(RED);
    printf("This is for testing purposes only. The output generated are not "
           "random. Don't use it for generating Random bits. !!!! \n");
    printf(RESET);
    printf("CTR DRBG - TEST \n");
    err = test_ctr_drbg();
    if (alcp_is_error(err)) {
        printf("Error in Testing CTR DRBG\n");
        return -1;
    }
    printf("\n");
    printf("HMAC DRBG - TEST \n");
    err = test_hmac_drbg();
    if (alcp_is_error(err)) {
        printf("Error in Testing HMAC DRBG\n");
        return -1;
    }
    printf("\n");

    return 0;
}
